from typing import Union

from ...zephyr_session import ZephyrSession


class TestExecutionEndpoints:
    """Api wrapper for "Test Execution" endpoints"""
    def __init__(self, session: ZephyrSession):
        self.session = session

    def get_test_executions(self, **kwargs):
        """
        Returns all test executions. Query parameters can be used to filter
        by project and folder
        """
        return self.session.get_paginated("testexecutions", params=kwargs)

    def create_test_execution(self,
                              project_key: str,
                              test_case_key: str,
                              test_cycle_key: str,
                              status_name: str,
                              **kwargs):
        """
        Creates a test execution. All required test execution custom fields
        should be present in the request
        """
        json = {"projectKey": project_key,
                "testCaseKey": test_case_key,
                "testCycleKey": test_cycle_key,
                "statusName": status_name}
        json.update(kwargs)
        return self.session.post("testexecutions", json=json)

    def get_test_execution(self, test_execution_id_or_key: Union[str, int], **kwargs):
        """Returns a test execution for the given ID"""
        return self.session.get(f"testexecutions/{test_execution_id_or_key}", params=kwargs)

    def get_links(self, test_execution_id_or_key: Union[str, int]):
        """Returns links for a test execution with specified ID"""
        return self.session.get(f"testexecutions/{test_execution_id_or_key}/links")

    def create_issue_links(self, test_execution_id_or_key: Union[str, int], issue_id: int):
        """Creates a link between a test execution and a Jira issue"""
        return self.session.post(f"testexecutions/{test_execution_id_or_key}/links",
                                 json={"issueId": issue_id})



from ...zephyr_session import ZephyrSession


class HealthcheckEndpoints:
    """Api wrapper for "Healthcheck" endpoints"""

    def __init__(self, session: ZephyrSession):
        self.session = session

    def get_health(self):
        """Check the health of this API

        :return: health check response body
        :rtype: dict
        """
        return self.session.get("healthcheck")



from ...zephyr_session import ZephyrSession


class FolderEndpoints:
    """Api wrapper for "Folder" endpoints"""

    def __init__(self, session: ZephyrSession):
        self.session = session

    def get_folders(self, **kwargs):
        """Returns all folders"""
        return self.session.get_paginated("folders", params=kwargs)

    def create_folder(self, name, project_key, folder_type, **kwargs):
        """Creates a folder"""
        json = {"name": name,
                "projectKey": project_key,
                "folderType": folder_type}
        json.update(kwargs)
        return self.session.post("folders", json=json)

    def get_folder(self, folder_id):
        """Returns a folder for the given ID"""
        return self.session.get(f"folders/{folder_id}")



from ...zephyr_session import ZephyrSession


class ProjectEndpoints:
    """Api wrapper for "Project" endpoints"""

    def __init__(self, session: ZephyrSession):
        self.session = session

    def get_projects(self):
        """Returns all projects"""
        return self.session.get_paginated("projects")

    def get_project(self, project_id_or_key):
        """Returns a project for the given ID or key"""
        return self.session.get(f"projects/{project_id_or_key}")



from .environments import EnvironmentEndpoints
from .test_cases import TestCaseEndpoints
from .projects import ProjectEndpoints
from .test_cycles import TestCycleEndpoints
from .test_plans import TestPlanEndpoints
from .test_executions import TestExecutionEndpoints
from .folders import FolderEndpoints
from .statuses import StatusEndpoints
from .priorities import PriorityEndpoints
from .links import LinkEndpoints
from .healthcheck import HealthcheckEndpoints
from .automations import AutomationEndpoints



from typing import Union

from ...zephyr_session import ZephyrSession


class TestCycleEndpoints:
    """Api wrapper for "Test Cycle" endpoints"""
    def __init__(self, session: ZephyrSession):
        self.session = session

    def get_all_test_cycles(self, **kwargs):
        """
        Returns all test cycles. Query parameters can be used to filter
        by project and folder
        """
        return self.session.get_paginated("testcycles", params=kwargs)

    def create_test_cycle(self, project_key: str, name: str, **kwargs):
        """
        Creates a Test Cycle. All required test cycle custom fields
        should be present in the request
        """
        json = {"projectKey": project_key,
                "name": name}
        json.update(kwargs)
        return self.session.post("testcycles", json=json)

    def get_test_cycle(self, test_cycle_id_or_key: Union[str, int]):
        """Returns a test cycle for the given key"""
        return self.session.get(f"testcycles/{test_cycle_id_or_key}")

    def update_test_cycle(self,
                          test_cycle_key: str,
                          test_cycle_id: int,
                          name: str,
                          project_id: str,
                          status_id: str,
                          **kwargs):
        """
        Updates an existing test cycle. If the project has test cycle
        custom fields, all custom fields should be present in the request.
        To leave any of them blank, please set them null
        if they are not required custom fields.
        """
        json = {"id": test_cycle_id,
                "key": test_cycle_key,
                "name": name,
                "project": {"id": project_id},
                "status": {"id": status_id}}
        json.update(kwargs)
        return self.session.put(f"testcycles/{test_cycle_key}", json=json)

    def get_links(self, test_cycle_id_or_key: Union[str, int]):
        """Returns links for a test cycle with specified key"""
        return self.session.get(f"testcycles/{test_cycle_id_or_key}/links")

    def create_issue_links(self, test_cycle_id_or_key: Union[str, int], issue_id: int):
        """Creates a link between a test cycle and a Jira issue"""
        json = {"issueId": issue_id}
        return self.session.post(f"testcycles/{test_cycle_id_or_key}/links/issues", json=json)

    def create_web_links(self, test_cycle_id_or_key: Union[str, int], url: str, **kwargs):
        """Creates a link between a test cycle and a generic URL"""
        json = {"url": url}
        json.update(kwargs)
        return self.session.post(f"testcycles/{test_cycle_id_or_key}/links/weblinks", json=json)



import logging
from urllib.parse import urlparse, parse_qs

from requests import Session


INIT_SESSION_MSG = "Initialize session by {}"


class ZephyrSession:
    """
    Zephyr Scale basic session object.

    :param base_url: url to make requests to
    :param token: auth token
    :param username: username
    :param password: password
    :param cookies: cookie dict
    """
    def __init__(self, base_url, token=None, username=None, password=None, cookies=None):
        self.base_url = base_url
        self._session = Session()
        self.logger = logging.getLogger(__name__)
        if token:
            self.logger.debug(INIT_SESSION_MSG.format("token"))
            self._session.headers.update({"Authorization": f"Bearer {token}"})
        elif username and password:
            self.logger.debug(INIT_SESSION_MSG.format("username and password"))
            self._session.auth = (username, password)
        elif cookies:
            self.logger.debug(INIT_SESSION_MSG.format("cookies"))
            self._session.cookies.update(cookies)
        else:
            raise Exception("Insufficient auth data")

    def _create_url(self, *args):
        """Helper for URL creation"""
        return self.base_url + "/".join(args)

    def _request(self, method: str, endpoint: str, return_raw: bool = False, **kwargs):
        """General request wrapper with logging and handling response"""
        self.logger.debug(f"{method.capitalize()} data: endpoint={endpoint} and {kwargs}")
        url = self._create_url(endpoint)
        response = self._session.request(method=method, url=url, **kwargs)
        if response.status_code < 400:
            if return_raw:
                return response
            if response.text:
                return response.json()
            return ""
        raise Exception(f"Error {response.status_code}. Response: {response.content}")

    def get(self, endpoint: str, params: dict = None, **kwargs):
        """Get request wrapper"""
        return self._request("get", endpoint, params=params, **kwargs)

    def post(self, endpoint: str, json: dict = None, **kwargs):
        """Post request wrapper"""
        return self._request("post", endpoint, json=json, **kwargs)

    def put(self, endpoint: str, json: dict = None, **kwargs):
        """Put request wrapper"""
        return self._request("put", endpoint, json=json, **kwargs)

    def delete(self, endpoint: str, **kwargs):
        """Delete request wrapper"""
        return self._request("delete", endpoint, **kwargs)

    def get_paginated(self, endpoint, params=None):
        """Get paginated data"""
        self.logger.debug(f"Get paginated data from endpoint={endpoint} and params={params}")
        if params is None:
            params = {}

        while True:
            response = self.get(endpoint, params=params)
            if "values" not in response:
                return
            for value in response.get("values", []):
                yield value
            if response.get("isLast") is True:
                break
            params_str = urlparse(response.get("next")).query
            params.update(parse_qs(params_str))
        return



from ...zephyr_session import ZephyrSession


class PriorityEndpoints:
    """Api wrapper for "Priority" endpoints"""

    def __init__(self, session: ZephyrSession):
        self.session = session

    def get_priorities(self, **kwargs):
        """Returns all priorities"""
        return self.session.get_paginated("priorities", params=kwargs)

    def get_priority(self, priority_id):
        """Returns a priority for the given ID"""
        return self.session.get(f"priorities/{priority_id}")



from ...zephyr_session import ZephyrSession


class StatusEndpoints:
    """Api wrapper for "Status" endpoints"""

    def __init__(self, session: ZephyrSession):
        self.session = session

    def get_statuses(self, **kwargs):
        """Returns all statuses"""
        return self.session.get_paginated("statuses", params=kwargs)

    def get_status(self, status_id):
        """Returns a status for the given ID"""
        return self.session.get(f"statuses/{status_id}")



import logging

from zephyr.scale.zephyr_session import ZephyrSession
from zephyr.scale.cloud.cloud_api import CloudApiWrapper


DEFAULT_BASE_URL = "https://api.zephyrscale.smartbear.com/v2/"

API_V2 = "v2"
API_V1 = "v1"


class ZephyrScale:
    """
    Zephyr Scale base object to interact with other objects or raw api by its methods.

    NOTE: Cloud API accepts only token auth, whereas Server API works with Jira auth methods.

    :param base_url: base API url to connect with
    :param api_version: 'v2' for Cloud and 'v1' for Server
    """
    def __init__(self, base_url=None, api_version=API_V2, **kwargs):
        base_url = DEFAULT_BASE_URL if not base_url else base_url
        session = ZephyrSession(base_url=base_url, **kwargs)

        if api_version.lower() == API_V2:
            self.api = CloudApiWrapper(session)
        elif api_version.lower() == API_V1:
            raise NotImplementedError("Server api wrappers have not been implemented yet.")
        else:
            raise ValueError("API version should be either 'v1' (Server) or 'v2' (Cloud)")
        self.logger = logging.getLogger(__name__)



from ...zephyr_session import ZephyrSession


class EnvironmentEndpoints:
    """Api wrapper for "Environment" endpoints"""

    def __init__(self, session: ZephyrSession):
        self.session = session

    def get_environments(self, **kwargs):
        """Returns all environments"""
        return self.session.get_paginated("environments", params=kwargs)

    def get_environment(self, environment_id):
        """Returns an environment for the given ID"""
        return self.session.get(f"environments/{environment_id}")



from ...zephyr_session import ZephyrSession


class LinkEndpoints:
    """Api wrapper for "Link" endpoints"""

    def __init__(self, session: ZephyrSession):
        self.session = session

    def delete_link(self, link_id):
        """Deletes a link for the given ID.

        :param link_id: The id of a link to delete
        :type link_id: int
        :return: response body
        :rtype: dict
        """
        return self.session.delete(f"links/{link_id}")



import logging

from zephyr.scale.zephyr_session import ZephyrSession
from zephyr.scale.cloud.endpoints import (AutomationEndpoints,
                                          EnvironmentEndpoints,
                                          FolderEndpoints,
                                          HealthcheckEndpoints,
                                          LinkEndpoints,
                                          PriorityEndpoints,
                                          ProjectEndpoints,
                                          StatusEndpoints,
                                          TestCaseEndpoints,
                                          TestCycleEndpoints,
                                          TestExecutionEndpoints,
                                          TestPlanEndpoints)


# pylint: disable=missing-function-docstring
class CloudApiWrapper:
    """Zephyr Scale Cloud Api wrapper. Contains wrappers by sections."""
    def __init__(self, session: ZephyrSession):
        self.session = session
        self.logger = logging.getLogger(__name__)

    @property
    def test_cases(self):
        return TestCaseEndpoints(self.session)

    @property
    def test_cycles(self):
        return TestCycleEndpoints(self.session)

    @property
    def test_plans(self):
        return TestPlanEndpoints(self.session)

    @property
    def test_executions(self):
        return TestExecutionEndpoints(self.session)

    @property
    def folders(self):
        return FolderEndpoints(self.session)

    @property
    def statuses(self):
        return StatusEndpoints(self.session)

    @property
    def priorities(self):
        return PriorityEndpoints(self.session)

    @property
    def environments(self):
        return EnvironmentEndpoints(self.session)

    @property
    def projects(self):
        return ProjectEndpoints(self.session)

    @property
    def links(self):
        return LinkEndpoints(self.session)

    @property
    def automations(self):
        return AutomationEndpoints(self.session)

    @property
    def healthcheck(self):
        return HealthcheckEndpoints(self.session)



from ...zephyr_session import ZephyrSession


class TestPlanEndpoints:
    """Api wrapper for "Test Plan" endpoints"""
    def __init__(self, session: ZephyrSession):
        self.session = session

    def get_test_plans(self, **kwargs):
        """Retrieves all test plans. Query parameters can be used to filter the results.

        Keyword arguments:
        :keyword projectKey: Jira project key filter
        :keyword maxResults: A hint as to the maximum number of results to return in each call
        :keyword startAt: Zero-indexed starting position. Should be a multiple of maxResults
        :return: dict with response body
        """
        return self.session.get_paginated("testplans", params=kwargs)



from ...zephyr_session import ZephyrSession


class AutomationEndpoints:
    """Api wrapper for "Automation" endpoints"""

    def __init__(self, session: ZephyrSession):
        self.session = session

    def post_custom_format(self):
        """
        Create results using Zephyr Scale's custom results format.
        """
        raise NotImplementedError

    def post_cucumber_format(self):
        """
        Create results using the Cucumber results format.
        """
        raise NotImplementedError

    def post_junit_xml_format(self):
        """
        Create results using the JUnit XML results format. Optionally,
        you can send a 'testCycle' part in your form data to customize
        the created test cycle.
        """
        raise NotImplementedError

    def get_testcases_zip(self, project_key):
        """
        Retrieve a zip file containing Cucumber Feature Files that matches
        the query passed as parameter.
        """
        params = {"projectKey": project_key}
        headers = {"Accept": "application/zip"}
        return self.session.get("automations/testcases",
                                return_raw=True,
                                params=params,
                                headers=headers)



from ...zephyr_session import ZephyrSession


class TestCaseEndpoints:
    """Api wrapper for "Test Case" endpoints"""
    def __init__(self, session: ZephyrSession):
        self.session = session

    def get_test_cases(self, **kwargs):
        """Retrieves all test cases. Query parameters can be used to filter the results.

        Keyword arguments:
        :keyword projectKey: Jira project key filter
        :keyword folderId: Folder ID filter
        :keyword maxResults: A hint as to the maximum number of results to return in each call
        :keyword startAt: Zero-indexed starting position. Should be a multiple of maxResults
        :return: dict with response body
        """
        return self.session.get_paginated("testcases", params=kwargs)

    def create_test_case(self, project_key: str, name: str, **kwargs):
        """Creates a test case. Fields priorityName and statusName will be set to
        default values if not informed. Default values are usually “Normal”
        for priorityName and “Draft” for statusName. All required test case custom
        fields should be present in the request.

        :param project_key: Jira project key
        :param name: test case name

        Keyword arguments:
        :keyword objective: A description of the objective
        :keyword precondition: Any conditions that need to be met
        :keyword estimatedTime: Estimated duration in milliseconds
        :keyword componentId: ID of a component from Jira
        :keyword priorityName: The priority name
        :keyword statusName: The status name
        :keyword folderId: ID of a folder to place the entity within
        :keyword ownerId: Atlassian Account ID of the Jira user
        :keyword labels: Array of labels associated to this entity
        :return: dict with response body
        """
        json = {"projectKey": project_key,
                "name": name}
        json.update(kwargs)
        return self.session.post("testcases", json=json)

    def get_test_case(self, test_case_key: str):
        """Returns a test case for the given key

        :param test_case_key: The key of the test case
        :return: dict with response body
        """
        return self.session.get(f"testcases/{test_case_key}")

    def update_test_case(self,
                         test_case_key: str,
                         test_case_id: int,
                         name: str,
                         project_id: int,
                         priority_id: int,
                         status_id: int,
                         **kwargs):
        """Updates an existing test case.

        :param test_case_key: The key of the test case
        :param test_case_id: integer id of the test
        :param name: test case nae
        :param project_id: project id
        :param priority_id: priority id
        :param status_id: status id

        :return: dict with response body
        """
        json = {"id": test_case_id,
                "key": test_case_key,
                "name": name,
                "project": {"id": project_id},
                "priority": {"id": priority_id},
                "status": {"id": status_id}}
        json.update(kwargs)
        return self.session.put(f"testcases/{test_case_key}", json=json)

    def get_links(self, test_case_key: str):
        """Returns links for a test case with specified key"""
        return self.session.get(f"testcases/{test_case_key}/links")

    def create_issue_links(self, test_case_key: str, issue_id: int):
        """Creates a link between a test case and a Jira issue"""
        json = {"issueId": issue_id}
        return self.session.post(f"testcases/{test_case_key}/links/issues", json=json)

    def create_web_links(self, test_case_key: str, url: str, **kwargs):
        """Creates a link between a test case and a generic URL"""
        json = {"url": url}
        json.update(kwargs)
        return self.session.post(f"testcases/{test_case_key}/links/weblinks", json=json)

    def get_versions(self, test_case_key: str, **kwargs):
        """
        Returns all test case versions for a test case with specified key.
        Response is ordered by most recent first
        """
        return self.session.get_paginated(f"testcases/{test_case_key}/versions", params=kwargs)

    def get_version(self, test_case_key: str, version: str):
        """Retrieves a specific version of a test case"""
        return self.session.get(f"testcases/{test_case_key}/versions/{version}")

    def get_test_script(self, test_case_key: str):
        """Returns the test script for the given test case"""
        return self.session.get(f"testcases/{test_case_key}/testscript")

    def create_test_script(self, test_case_key: str, script_type: str, text: str):
        """
        Creates or updates the test script for a test case. If the test case currently
        has a sequence of test steps assigned to it, these will be implicitly removed.
        """
        json = {"type": script_type, "text": text}
        return self.session.post(f"testcases/{test_case_key}/testscript", json=json)

    def get_test_steps(self, test_case_key: str, **kwargs):
        """
        Returns the test steps for the given test case. Provides a paged response,
        with 100 items per page.
        """
        return self.session.get_paginated(f"testcases/{test_case_key}/teststeps", params=kwargs)

    def post_test_steps(self, test_case_key: str, mode: str, items: list):
        """
        Assigns a series of test steps to a test case, appending them to any existing
        sequence of test steps. A maximum of 100 steps can be posted per request. Consumers
        should not attempt to parallelize this operation, as the order of the steps is defined
        by the input order. If this endpoint is called on a test case that already has
        a plain text or BDD test script, that test script will implicitly be removed.
        All required step custom fields should be present in the request.
        """
        json = {"mode": mode,
                "items": items}
        return self.session.post(f"testcases/{test_case_key}/teststeps", json=json)
